@@ -7,11 +7,13 @@ import json
 import logging
 import os
 import sys
+import threading
 import yaml
+import concurrent.futures
 
 from .config import merge_cfgs
 from .images import get_is_configs, import_images
-from .utils import load_cfg_file, oc, trigger_builds, wait_for_ready_threaded
+from .utils import apply_template, load_cfg_file, trigger_builds, wait_for_ready_threaded
 from .secrets import import_secrets, SecretImporter
 from .templates import Template, get_templates_in_dir
 
@@ -78,10 +80,10 @@ def deploy_components(
         processed_templates_by_name[comp_name] = template
 
         log.info("Deploying component '%s'", comp_name)
-        oc("apply", "-f", "-", "-n", project_name, _in=template.dump_processed_json())
+        apply_template(project_name, template)
 
         # Mark certain resources in this component as ones we need to wait on
-        for restype in ("dc", "sts"):
+        for restype in ("deployment", "deploymentconfig", "statefulset", "daemonset"):
             resources_to_wait_for.extend(
                 [(restype, name) for name in template.get_processed_names_for_restype(restype)]
             )
@@ -213,7 +215,7 @@ def _get_custom_deploy_methods(service_set, service_set_dir, root_custom_dir):
         )
 
     if not module:
-        module = _load_module(os.path.join(root_custom_dir, f"deploy.py"), service_set)
+        module = _load_module(os.path.join(root_custom_dir, "deploy.py"), service_set)
 
     if not module:
         log.exception("Error loading custom deploy script, using default deploy methods")
@@ -322,6 +324,8 @@ class DeployRunner(object):
         skip=None,
         dry_run=False,
         dry_run_opts=None,
+        concurrent=False,
+        threadpool_size=os.cpu_count(),
     ):
         self.template_dir = template_dir
         self.root_custom_dir = root_custom_dir
@@ -338,6 +342,8 @@ class DeployRunner(object):
         self.dry_run_opts = dry_run_opts or {}
         self.env_config_handler = env_config_handler
         self._base_is_configs = {}
+        self.concurrent = concurrent
+        self.threadpool_size = threadpool_size
 
     def _get_variables(self, service_set_name, service_set_dir, component):
         variables = {}
@@ -482,7 +488,11 @@ class DeployRunner(object):
         return merge_cfgs(set_cfg, set_env_cfg)
 
     def _deploy_service_set(self, service_set):
+        if self.concurrent:
+            threading.current_thread().name = service_set[:21]
+
         log.info("Handling config for service set '%s'", service_set)
+
         processed_templates = {}
 
         dir_path = os.path.join(self.template_dir, service_set)
@@ -553,6 +563,35 @@ class DeployRunner(object):
         self._deployed_service_sets.append(service_set)
         return processed_templates
 
+    def _deploy_sets_concurrently(self, service_sets_selected):
+        all_processed_templates = {}
+        failed_sets = []
+
+        log.info(
+            "deploying service sets (%s) in this stage concurrently, threadpool size: %d",
+            ", ".join(service_sets_selected),
+            self.threadpool_size,
+        )
+
+        with concurrent.futures.ThreadPoolExecutor(max_workers=self.threadpool_size) as executor:
+            futures = {
+                executor.submit(self._deploy_service_set, service_set): service_set
+                for service_set in service_sets_selected
+            }
+            for future in concurrent.futures.as_completed(futures):
+                service_set = futures[future]
+                try:
+                    all_processed_templates[service_set] = future.result()
+                except Exception:
+                    log.exception("Service set '%s' hit exception", service_set)
+                    failed_sets.append(service_set)
+
+        if failed_sets:
+            log.error("deploys failed for service set(s): %s", ", ".join(failed_sets))
+            raise Exception("deploys failed in stage")
+
+        return all_processed_templates
+
     def run(self):
         """
         Load the "_cfg.yml" at {dir_path}, parse it, and deploy the defined components
@@ -592,14 +631,27 @@ class DeployRunner(object):
 
         for stage in sorted(deploy_order.keys()):
             service_sets = deploy_order[stage].get("components", [])
+            service_sets_selected = []
             for service_set in service_sets:
                 if service_set not in sets_for_deploy:
                     log.info(
                         "Skipping service set '%s', not selected for deploy at runtime", service_set
                     )
                     continue
-                else:
+                service_sets_selected.append(service_set)
+
+            if self.concurrent:
+                all_processed_templates.update(
+                    self._deploy_sets_concurrently(service_sets_selected)
+                )
+            else:
+                for service_set in service_sets_selected:
                     all_processed_templates[service_set] = self._deploy_service_set(service_set)
 
+        total_components = sum([len(templates) for _, templates in all_processed_templates.items()])
+        log.info("Processed templates for %d total components", total_components)
+        if not total_components:
+            raise ValueError("The chosen combination of '-s' or '-p' did not match any components")
+
         if self.dry_run:
             generate_dry_run_content(all_processed_templates, **self.dry_run_opts)
