@@ -8,11 +8,12 @@ import threading
 import time
 import os
 import yaml
+import re
 from functools import reduce
 
 from anytree import Node, RenderTree, PreOrderIter
 import sh
-from sh import ErrorReturnCode
+from sh import ErrorReturnCode, TimeoutException
 from wait_for import wait_for, TimedOutError
 
 log = logging.getLogger(__name__)
@@ -20,8 +21,12 @@ log = logging.getLogger(__name__)
 # Resource types and their cli shortcuts
 # Mostly listed here: https://docs.openshift.com/online/cli_reference/basic_cli_operations.html
 SHORTCUTS = {
+    "configmap": "cm",
+    "cronjob": "cj",
     "build": None,
     "buildconfig": "bc",
+    "daemonset": "ds",
+    "deployment": "deploy",
     "deploymentconfig": "dc",
     "event": "ev",
     "imagestream": "is",
@@ -29,7 +34,7 @@ SHORTCUTS = {
     "imagestreamimage": "isimage",
     "job": None,
     "limitrange": "limits",
-    "node": None,
+    "node": "no",
     "pod": "po",
     "resourcequota": "quota",
     "replicationcontroller": "rc",
@@ -39,20 +44,30 @@ SHORTCUTS = {
     "statefulset": "sts",
     "persistentvolume": "pv",
     "persistentvolumeclaim": "pvc",
-    "configmap": None,
+    "replicaset": "rs",
     "route": None,
 }
 
 
-def validate_list_of_strs(item_name, section, l):
+INVALID_RESOURCE_REGEX = re.compile(
+    r'The (\S+) "(\S+)" is invalid: metadata.resourceVersion: Invalid value: 0x0'
+)
+
+
+def abort():
+    log.error("Hit fatal error!  Aborting.")
+    sys.exit(1)
+
+
+def validate_list_of_strs(item_name, section, list_):
     bad = False
 
     try:
-        iter(l)
+        iter(list_)
     except TypeError:
         bad = True
     else:
-        if not all([isinstance(i, str) for i in l]):
+        if not all([isinstance(i, str) for i in list_]):
             bad = True
 
     if bad:
@@ -145,10 +160,10 @@ def get_dir(value, default_value, dir_type, optional=False):
     path_exists_but_not_a_dir = os.path.exists(path) and not os.path.isdir(path)
     if required_dir_does_not_exist:
         log.error("%s directory missing: %s", dir_type, path)
-        sys.exit(1)
+        abort()
     if required_dir_is_not_a_dir or path_exists_but_not_a_dir:
         log.error("%s directory invalid: %s", dir_type, path)
-        sys.exit(1)
+        abort()
     path = os.path.abspath(path)
     if os.path.exists(path):
         log.info("Found %s path: %s", dir_type, path)
@@ -163,13 +178,13 @@ def all_sets(template_dir):
             cfg_data = load_cfg_file(f"{template_dir}/_cfg.yml")
         except ValueError as err:
             log.error("Error: template dir '%s' invalid: %s", template_dir, str(err))
-            sys.exit(1)
+            abort()
 
     try:
         stages = cfg_data["deploy_order"]
     except KeyError:
         log.error("Error: template dir '%s' invalid: _cfg file has no 'deploy_order'")
-        sys.exit(1)
+        abort()
 
     sets = reduce(lambda acc, s: acc + s.get("components", []), stages.values(), [])
 
@@ -206,37 +221,53 @@ def _exec_oc(*args, **kwargs):
     _silent = kwargs.pop("_silent", False)
     _ignore_immutable = kwargs.pop("_ignore_immutable", True)
     _retry_conflicts = kwargs.pop("_retry_conflicts", True)
+    _stdout_log_prefix = kwargs.pop("_stdout_log_prefix", " |stdout| ")
+    _stderr_log_prefix = kwargs.pop("_stderr_log_prefix", " |stderr| ")
 
-    kwargs["_bg_exc"] = True
+    kwargs["_bg"] = True
+    kwargs["_bg_exc"] = False
 
     err_lines = []
     out_lines = []
 
-    def _err_line_handler(line):
-        log.info(" |stderr| %s", line.rstrip())
+    def _err_line_handler(line, _, process):
+        threading.current_thread().name = f"pid-{process.pid}"
+        log.info("%s%s", _stderr_log_prefix, line.rstrip())
         err_lines.append(line)
 
-    def _out_line_handler(line):
+    def _out_line_handler(line, _, process):
+        threading.current_thread().name = f"pid-{process.pid}"
         if not _silent:
-            log.info(" |stdout| %s", line.rstrip())
+            log.info("%s%s", _stdout_log_prefix, line.rstrip())
         out_lines.append(line)
 
     retries = 3
     last_err = None
     for count in range(1, retries + 1):
+        cmd = sh.oc(*args, **kwargs, _tee=True, _out=_out_line_handler, _err=_err_line_handler)
+        if not _silent:
+            cmd_args, cmd_kwargs = _get_logging_args(args, kwargs)
+            log.info("running (pid %d): oc %s %s", cmd.pid, cmd_args, cmd_kwargs)
         try:
-            if not _silent:
-                cmd_args, cmd_kwargs = _get_logging_args(args, kwargs)
-                log.info("Running command: oc %s %s", cmd_args, cmd_kwargs)
-            return sh.oc(
-                *args, **kwargs, _tee=True, _out=_out_line_handler, _err=_err_line_handler
-            ).wait()
+            return cmd.wait()
         except ErrorReturnCode as err:
             # Sometimes stdout/stderr is empty in the exception even though we appended
             # data in the callback. Perhaps buffers are not being flushed ... so just
-            # set the out lines/err lines we captured on the Exception before re-raising it
+            # set the out lines/err lines we captured on the Exception before re-raising it by
+            # re-init'ing the err and causing it to rebuild its message template.
+            #
+            # see https://github.com/amoffat/sh/blob/master/sh.py#L381
+            err.__init__(
+                full_cmd=err.full_cmd,
+                stdout="\n".join(out_lines).encode(),
+                stderr="\n".join(err_lines).encode(),
+                truncate=err.truncate,
+            )
+
+            # Make these plain strings for easier exception handling
             err.stdout = "\n".join(out_lines)
             err.stderr = "\n".join(err_lines)
+
             last_err = err
 
             # Ignore warnings that are printed to stderr in our error analysis
@@ -270,6 +301,8 @@ def oc(*args, **kwargs):
         _silent: don't print command or resulting stdout (default False)
         _ignore_immutable: ignore errors related to immutable objects (default True)
         _retry_conflicts: retry commands if a conflict error is hit
+        _stdout_log_prefix: prefix this string to stdout log output (default " |stdout| ")
+        _stderr_log_prefix: prefix this string to stderr log output (default " |stderr| ")
 
     Returns:
         None if cmd fails and _exit_on_err is False
@@ -285,13 +318,37 @@ def oc(*args, **kwargs):
         if _reraise:
             raise
         elif _exit_on_err:
-            log.error("Command failed!  Aborting.")
-            sys.exit(1)
+            abort()
         else:
             if not kwargs.get("_silent"):
                 log.warning("Non-zero return code ignored")
 
 
+def apply_template(project, template):
+    try:
+        oc("apply", "-f", "-", "-n", project, _in=template.dump_processed_json(), _reraise=True)
+    except ErrorReturnCode as err:
+        # Work-around for resourceVersion errors.
+        # See https://www.timcosta.io/kubernetes-service-invalid-clusterip-or-resourceversion/
+        matches = INVALID_RESOURCE_REGEX.findall(err.stderr)
+        if matches:
+            for restype, name in matches:
+                restype = restype.rstrip("s")  # remove plural language
+                if template.get_processed_item(restype, name):  # ensure we sent this item's config
+                    log.warning(
+                        "Removing last-applied-configuration annotation from %s/%s", restype, name
+                    )
+                    oc(
+                        "annotate",
+                        restype,
+                        name,
+                        "kubectl.kubernetes.io/last-applied-configuration-",
+                    )
+            oc("apply", "-f", "-", "-n", project, _in=template.dump_processed_json())
+        else:
+            abort()
+
+
 def switch_to_project(project):
     try:
         oc("get", "project", project, _reraise=True)
@@ -396,7 +453,7 @@ def _check_status_for_restype(restype, json_data):
 
     restype = parse_restype(restype)
 
-    if restype == "deploymentconfig":
+    if restype == "deploymentconfig" or restype == "deployment":
         spec_replicas = json_data["spec"]["replicas"]
         available_replicas = status.get("availableReplicas", 0)
         updated_replicas = status.get("updatedReplicas", 0)
@@ -410,8 +467,13 @@ def _check_status_for_restype(restype, json_data):
         ready_replicas = status.get("readyReplicas", 0)
         return ready_replicas == spec_replicas
 
+    elif restype == "daemonset":
+        desired = status.get("desiredNumberScheduled", 1)
+        available = status.get("numberAvailable")
+        return desired == available
+
     elif restype == "pod":
-        if status.get("phase") == "Running":
+        if status.get("phase").lower() == "running":
             return True
 
     elif restype == "build":
@@ -441,6 +503,41 @@ def _check_status_for_restype(restype, json_data):
         )
 
 
+def _wait_with_periodic_status_check(timeout, key, restype, name):
+    """Check if resource is ready using _check_status_for_restype, periodically log an update."""
+    time_last_logged = time.time()
+    time_remaining = timeout
+
+    def _ready():
+        nonlocal time_last_logged, time_remaining
+
+        j = get_json(restype, name)
+        if _check_status_for_restype(restype, j):
+            return True
+
+        if time.time() > time_last_logged + 60:
+            time_remaining -= 60
+            if time_remaining:
+                log.info("[%s] waiting %dsec longer", key, time_remaining)
+                time_last_logged = time.time()
+        return False
+
+    wait_for(
+        _ready, timeout=timeout, delay=5, message="wait for '{}' to be ready".format(key),
+    )
+
+
+def wait_for_exists(restype, name, timeout=300):
+    restype = parse_restype(restype)
+    key = "{}/{}".format(SHORTCUTS.get(restype) or restype, name)
+    log.info("[%s] waiting up to %dsec for resource to exist", key, timeout)
+
+    def _exists():
+        return get_json(restype, name) is not {}
+
+    wait_for(_exists, timeout=timeout, delay=5, message="wait for '{}' to exist".format(key))
+
+
 def wait_for_ready(restype, name, timeout=300, exit_on_err=False, _result_dict=None):
     """
     Wait {timeout} for resource to be complete/ready/active.
@@ -459,36 +556,42 @@ def wait_for_ready(restype, name, timeout=300, exit_on_err=False, _result_dict=N
     to store the result of this wait as:
         _result_dict[resource_name] = True or False
     """
-    key = "{}/{}".format(restype, name)
+    restype = parse_restype(restype)
+    key = "{}/{}".format(SHORTCUTS.get(restype) or restype, name)
 
     if _result_dict is None:
         _result_dict = dict()
     _result_dict[key] = False
 
-    log.info("Waiting up to %dsec for '%s' to complete", timeout, key)
-
-    def _complete():
-        j = get_json(restype, name)
-        if _check_status_for_restype(restype, j):
-            _result_dict[key] = True
-            log.info("'%s' is ready!", key)
-            return True
-        return False
+    log.info("[%s] waiting up to %dsec for resource to be ready", key, timeout)
 
     try:
-        wait_for(
-            _complete,
-            timeout=timeout,
-            delay=5,
-            message="wait for '{}' to complete".format(key),
-            log_on_loop=True,
-        )
+        # Do not use rollout status for statefulset/daemonset yet until we can handle
+        # https://github.com/kubernetes/kubernetes/issues/64500
+        if restype in ["deployment", "deploymentconfig"]:
+            # use oc rollout status for the applicable resource types
+            oc(
+                "rollout",
+                "status",
+                key,
+                _reraise=True,
+                _timeout=timeout,
+                _stdout_log_prefix=f"[{key}] ",
+                _stderr_log_prefix=f"[{key}]  ",
+            )
+        else:
+            _wait_with_periodic_status_check(timeout, key, restype, name)
+
+        log.info("[%s] is ready!", key)
+        _result_dict[key] = True
         return True
-    except (TimedOutError, StatusError):
-        log.exception("Hit error waiting on '%s'", key)
-        if exit_on_err:
-            sys.exit(1)
-        return False
+    except (StatusError, ErrorReturnCode) as err:
+        log.error("[%s] hit error waiting for resource to be ready: %s", key, str(err))
+    except (TimeoutException, TimedOutError):
+        log.error("[%s] timed out waiting for resource to be ready", key)
+    if exit_on_err:
+        abort()
+    return False
 
 
 def wait_for_ready_threaded(restype_name_list, timeout=300, exit_on_err=False):
@@ -510,6 +613,8 @@ def wait_for_ready_threaded(restype_name_list, timeout=300, exit_on_err=False):
         for restype, name in restype_name_list
     ]
     for thread in threads:
+        thread.daemon = True
+        thread.name = thread.name.lower()  # because I'm picky
         thread.start()
     for thread in threads:
         thread.join()
@@ -519,7 +624,7 @@ def wait_for_ready_threaded(restype_name_list, timeout=300, exit_on_err=False):
     if failed:
         log.info("Some resources failed to become ready: %s", ", ".join(failed))
         if exit_on_err:
-            sys.exit(1)
+            abort()
         return False
     return True
 
@@ -556,7 +661,7 @@ def no_pods_running(dc_name):
     """
     Return true if there are no pods running in the deployment
     """
-    return not all_pods_running(dc_name)
+    return not any_pods_running(dc_name)
 
 
 def stop_deployment(dc_name, timeout=180):
